"""Delta upgrade test."""

import logging
import time
import unittest

from upgrade.device import Device
from upgrade.environment import CLEANUP
from upgrade.keyring import Keyrings
from upgrade.upgrade_files import UpgradeFiles


class DeltaUpgradeTestCase(unittest.TestCase):
    def setUp(self):
        logging.basicConfig(
            level=logging.DEBUG,
            format='%(asctime)s %(levelname)s: %(message)s')
        logging.info('Test case started')

    def tearDown(self):
        CLEANUP.run()
        logging.info('Test case finished')

    def test_delta(self):
        keyrings = Keyrings()

        device = Device()
        device.flash()
        device.flash_recovery_partition(keyrings['archive-master'])

        # Push upgrade files after recovery partition is flashed
        # (otherwise files are removed after reboot)
        upgrade_files = UpgradeFiles(device, keyrings)
        upgrade_files.push()

        # Launch services
        logging.info('Launching HTTP/HTTPS server...')
        server = device.run(
            '/tmp/upgrade-test/scripts/server.py 2>&1 '
            '| tee /var/log/system-image/test-server.log')

        logging.info('Launching client DBus service...')
        dbus_service = device.run(
            'system-image-dbus '
            '-C /tmp/upgrade-test/data/client.ini '
            '--testing=live '
            '-v 2>&1 '
            '| tee /var/log/system-image/test-client.log')

        # Check services have been correctly launched
        server.wait_for('server in')
        server.wait_for('server in')
        logging.info('HTTP/HTTPS server running')

        dbus_service.wait_for('SystemImage dbus main loop started')
        logging.info('Client DBus service running')

        logging.info('Using DBus API to get upgrade from server...')
        dbus_client = device.run(
            '/tmp/upgrade-test/scripts/check-updates.py 2>&1 '
            '| tee /var/log/system-image/test-update.log')

        # Keep track of progress
        dbus_client.wait_for('Update available')
        logging.info('Update available')
        dbus_client.wait_for('Calling DownloadUpdate')
        logging.info('Downloading update')
        dbus_client.wait_for('Update downloaded')
        logging.info('Update downloaded')
        dbus_client.wait_for('Calling ApplyUpdate...')
        logging.info('Applying update')

        # Verify content of /android/cache/recovery
        dbus_service.wait_for('all files available')
        ls = device.run('ls /android/cache/recovery/ubuntu_command')
        self.assertNotIn(
            'No such file or directory', ls.stdout,
            'ubuntu_command file not found')
        dbus_service.wait_for('rebooting')

        # Terminate processes
        logging.info('Terminating processes...')
        dbus_client.process.terminate()
        dbus_service.process.terminate()
        server.process.terminate()

        # TBD: Investigate why device isn't rebooting on its own for some
        # reason
        logging.info('Rebooting device')
        device.run('/sbin/reboot -f recovery')

        logging.info('Waiting for device to reboot...')
        time.sleep(60)
        device.wait()

        logging.info('Checking that upgrade was correctly applied')
        cat = device.run('cat /etc/upgrade-test').wait()
        self.assertEqual(cat.stdout, 'Upgrade test')

        logging.info('Success!')
