"""Run shell commands.

This module provides a class that is a wrapper around the subprocess.Popen
class.

"""

import concurrent.futures
import logging
import re
import subprocess


class Command(object):

    """A wrapper around subprocess.Popen with other defaults and logging.

    The different set of defaults being applied are:
        - stdout captured
        - stderr redirected to stdout
        - command executed through as shell

    :param cmd: Command that will be executed
    :type cmd: str
    :param kwargs: Keyword arguments to pass to the subprocess.Popen object
    :type kwargs: dict

    """

    TIMEOUT = 30

    def __init__(self, cmd, **kwargs):
        self.process = None
        self.cmd = cmd
        if isinstance(self.cmd, str):
            self.cmd_str = self.cmd
        elif isinstance(self.cmd, list):
            self.cmd_str = ' '.join(self.cmd)
        else:
            raise TypeError('cmd should be either <str> or <list>')

        # Output will be captured in an array and then joined into a string
        self.stdout = []

        kwargs.setdefault('stdout', subprocess.PIPE)
        kwargs.setdefault('stderr', subprocess.STDOUT)
        kwargs.setdefault('shell', True)
        self.kwargs = kwargs

        self.process = subprocess.Popen(self.cmd, **self.kwargs)
        logging.debug('[{}] Running {!r}...'
                      .format(self.process.pid, self.cmd_str))

    def communicate(self):
        """Wait until command completes, write logs afterwards.

        This method wraps the Popen.communicate method to write results to logs

        :returns: The command object (useful to chain multiple calls)
        :rtype: upgrade.command.Command

        """
        stdout, _ = self.process.communicate()
        self.stdout = stdout.decode('utf-8') if stdout else None
        if self.process.returncode == 0:
            log_level = logging.DEBUG
        else:
            log_level = logging.ERROR

        message = ['Return code: {}'.format(self.process.returncode)]
        if self.stdout:
            message.append('stdout:\n{}'.format(self.stdout))
        message = '\n'.join(message)

        logging.log(log_level, message)
        return self

    def wait(self, timeout=TIMEOUT):
        """Wait until command completes, writing logs as when possible.

        This method is a wrapper around the wait_for method to write every
        output line to logs and then log the process return code as well

        :param timeout: Timeout in seconds to wait for the command to finish
        :type timeout: int
        :returns: The command object (useful to chain multiple calls)
        :rtype: upgrade.command.Command

        """
        return self.wait_for(None, timeout)

    def wait_for(self, pattern=None, timeout=TIMEOUT):
        """Wait until the given pattern is found.

        This method checks stdout line by line until a line that matches the
        given pattern is found. Note that the command is chainable and can be
        used multiple times.

        The pattern is compiled as a regular expression and the matching method
        used to match the expression is search. Note that start/end of string
        characters have to be used if matching substrings isn't desired. Use
        None to avoid pattern matching and wait until EOF.

        :param pattern: Pattern that should be matched against the output lines
        :type patern: str | None
        :param timeout: Timeout in seconds to wait for the pattern to match
        :type timeout: int
        :returns: The command object (useful to chain multiple calls)
        :rtype: upgrade.command.Command

        """
        assert self.process
        expression = re.compile(pattern) if pattern else None

        def target():
            while True:
                line = self.process.stdout.readline().decode('utf-8')
                if not line:
                    # Raise exception if pattern wasn't intended to match EOF
                    if pattern is not None:
                        error_msg = ('[{}] Pattern {!r} not found'
                                     .format(self.process.pid, pattern))
                        logging.error(error_msg)
                        raise EOFError(error_msg)

                    # Write return code to logs
                    assert self.process.poll() is not None
                    self.stdout = ''.join(self.stdout)

                    if self.process.returncode == 0:
                        log_level = logging.DEBUG
                    else:
                        log_level = logging.ERROR
                    logging.log(
                        log_level,
                        '[{}] Return code: {}'
                        .format(self.process.pid, self.process.returncode))
                    return self

                self.stdout.append(line)
                logging.debug('[{}] {}'
                              .format(self.process.pid, line.rstrip()))

                if expression and expression.search(line):
                    logging.debug('[{}] Pattern {!r} found'
                                  .format(self.process.pid, pattern))
                    break

        with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
            future = executor.submit(target)
            future.result(timeout)

        return self
