"""Interact with the device using a pythonic interface."""
import fnmatch
import logging
import os
import shutil
import tarfile
import tempfile
import time

from os.path import join, basename

from upgrade.command import Command
from upgrade.image import get_download_dir
from upgrade.environment import CLEANUP


class Device(object):

    """Device object to interact with the real device.

    The device is expected to be connected to the hardware in which the
    test case runs, so that commands can be executed though adb.

    """

    def run(self, cmd):
        """Run command in device through adb.

        This is a wrapper around upgrade.command.Command to prepend `adb shell`
        to the command string passed.

        :param cmd: Command to execute in device (without `adb shell` prefix)
        :type cmd: str
        :returns: Command object after running it (in the background or not)
        :rtype: upgrade.command.Command

        """
        return Command("adb shell '{}'".format(cmd))

    def wait(self, timeout=60):
        """Wait for device to be available.

        This is a wraper around `adb wait-for-device` with and additonal sleep
        time to make sure device is really available.

        :param timeout: Addional time in seconds to wait for the device
        :type timeout: int
        :returns: Command object after the execution of `adb wait-for-device`
        :rtype: upgrade.device.Device

        """
        cmd = Command('adb wait-for-device').wait()
        time.sleep(timeout)
        return cmd

    def push(self, source, destination):
        """Push files to device.

        :param source: Path to source file/directory
        :type source: str
        :param destination: Path to destination directory
        :type destination: str

        """
        # Ensure destination is a file name when needed
        if not os.path.isdir(source):
            destination = join(destination, basename(source))

        logging.info('Pushing {} to {}...'.format(source, destination))
        Command('adb push {} {}'.format(source, destination)).wait()

    def get_info(self):
        """Get device installed system image information.

        :returns: The upgrade channel and the device model
        :rtype: tuple(str, str)

        """
        logging.info('Getting image info...')
        cmd = self.run('system-image-cli -i').wait()
        lines = cmd.stdout.splitlines()
        info = dict([line.split(': ', 1) for line in lines])
        channel = info['channel']
        device = info['device name']
        logging.debug('channel: {}, device: {}'.format(channel, device))
        return channel, device

    def flash(self, timeout=1000):
        """Flash device with latest image.

        :param timeout: Timeout in seconds to flash the device
        :type timeout: int

        """
        logging.info('Flashing device with latest image '
                     '(this might take a while)...')
        Command('phablet-flash ubuntu-system '
                '--channel devel-proposed --bootstrap').communicate()
        logging.info('Waiting for device to boot up...')
        self.wait()
        logging.info('Configuring network interface...')
        Command('phablet-network').wait()
        time.sleep(60)

        logging.info('Making sure image is writable...')
        self.run('touch /userdata/.writable_image').wait()
        logging.info('Rebooting device...')
        self.run('reboot').wait()
        logging.info('Waiting for device to boot up...')
        self.wait(120)
        logging.info('Installing system-image-dev package...')
        self.run('apt-get install system-image-dev').wait()
        apt_cache = self.run(
            'apt-cache policy system-image-dev | grep Installed').wait()
        assert '(none)' not in apt_cache.stdout

    def flash_recovery_partition(self, archive_master_keyring):
        """Flash recovery partition with test keys."""
        logging.info('Patching recovery image to use testing keys...')
        download_dir = get_download_dir(
            os.path.join('phablet-flash', 'imageupdates', 'pool'))
        logging.debug('Looking for device files in {}...'.format(download_dir))
        files = os.listdir(download_dir)
        device_files = sorted(
            fnmatch.filter(files, 'device-*.tar.xz'),
            key=lambda f: os.path.getmtime(join(download_dir, f)),
            reverse=True,
        )
        device_file = join(download_dir, device_files[0])
        logging.debug('Device file: {}'.format(device_file))

        recovery_dir = tempfile.mkdtemp(
            prefix='upgrade-test-recovery-partition-')
        CLEANUP.add_directory(recovery_dir)

        # Locate latest image file downloaded by phablet flash
        recovery_fname = 'recovery.img'
        with tarfile.open(device_file) as tf:
            logging.debug('Extracting recovery.img...')
            tf.extract(join('partitions', recovery_fname), recovery_dir)

        recovery_path = join(recovery_dir, 'partitions', recovery_fname)

        logging.debug('Extracting initrd...')
        Command(
            'abootimg -x {} cmdline kernel initrd'.format(recovery_path),
            cwd=recovery_dir,
        ).wait()

        logging.debug('Updating image size...')
        Command(
            "sed -ri 's/bootsize = 0x[0-9]+/bootsize = 0xa00000/' cmdline",
            cwd=recovery_dir,
        ).wait()

        logging.debug('Extracting archive master keyring...')
        cpio_dir = join(recovery_dir, 'cpio')
        os.mkdir(cpio_dir)
        Command(
            'cat ../initrd | gzip -d | cpio -i',
            cwd=cpio_dir,
        ).wait()

        logging.debug('Copying *testing* archive master keyring...')
        target_dir = join(cpio_dir, 'etc', 'system-image')
        shutil.copy(archive_master_keyring.tar_path, target_dir)
        shutil.copy(archive_master_keyring.asc_path, target_dir)

        logging.debug('Repacking initrd...')
        with open(join(recovery_dir, 'initrd'), 'wb') as fd:
            Command(
                'find . | cpio -o --format=newc | gzip -9',
                stdout=fd,
                cwd=cpio_dir,
            ).communicate()

        logging.debug('Repacking recovery.img...')
        Command(
            'abootimg -u {} -f cmdline'.format(recovery_path),
            cwd=recovery_dir,
        ).wait()
        Command(
            'abootimg -u {} -r initrd'.format(recovery_path),
            cwd=recovery_dir,
        ).wait()

        logging.info('Flashing test recovery image...')
        self.run('reboot -f bootloader').wait()
        Command('fastboot flash recovery {}'.format(recovery_path)).wait()
        Command('fastboot reboot').wait()

        logging.info('Waiting for device to boot up...')
        self.wait(90)
