"""Keyring generation and signing code."""

import gnupg
import json
import logging
import shutil
import tarfile
import tempfile

from os.path import (
    join,
    basename,
)

from upgrade.environment import CLEANUP, JSON_OPTIONS


class Keyrings(object):

    """Create all keyrings and signature files properly."""

    def __init__(self):
        # Access keyrings by their type
        self.keyring = {}

        # Archive master keyring
        amk = Keyring('archive-master')
        amk.sign_keyring(amk)

        # Image master keyring
        imk = Keyring('image-master')
        amk.sign_keyring(imk)

        # Image signing keyring
        isk = Keyring('image-signing')
        imk.sign_keyring(isk)

        # Keep track of keyring objects to sign other files
        self.keyring['archive-master'] = amk
        self.keyring['image-master'] = imk
        self.keyring['image-signing'] = isk

    def __getitem__(self, keyring_type):
        """Access keyring by type.

        :param keyring_type: Keyring type
        :type keyring_type: str
        :returns: Keyring object for the given type
        :rtype: upgrade.keyring.Keyring

        """
        return self.keyring[keyring_type]

    def export(self, directory):
        """Export keyrings files to a given directory.

        :param directory: Path to the directory to write keyring files
        :type directory: str

        """
        for keyring in self.keyring.values():
            keyring.export(directory)


class Keyring(object):

    """Create keyring files as used by the upgrade components.

    :param keyring_type:
        The type of keyring to create. This is metadata that will be added to
        the json file.
    :type keyring_type: str

    """

    KEYRING_FNAME = 'keyring.gpg'
    KEYRING_JSON_FNAME = 'keyring.json'

    def __init__(self, keyring_type):
        self.keyring_type = keyring_type
        self.tar_path = None
        self.asc_path = None

        logging.debug('Creating {} keyring directory...'.format(keyring_type))
        self.keyring_dir = tempfile.mkdtemp(prefix='upgrade-test-keyring-')
        CLEANUP.add_directory(self.keyring_dir)

        logging.debug('Creating {} keyring file...'.format(keyring_type))
        self.gpg = gnupg.GPG(
            gnupghome=self.keyring_dir,
            keyring=self.KEYRING_FNAME)

        logging.debug('Creating {} keyring json file...'.format(keyring_type))
        keyring_json_path = self.get_abspath(self.KEYRING_JSON_FNAME)
        with open(keyring_json_path, 'w') as fp:
            json.dump({'type': keyring_type}, fp, **JSON_OPTIONS)

        logging.info('Generating {} key (this might take a while)...'
                     .format(keyring_type))
        self.key = self.gpg.gen_key(self.gpg.gen_key_input())

        self._create_tar_file()

    def _create_tar_file(self):
        logging.debug('Creating {} tar file..'.format(self.keyring_type))
        keyring_path = self.get_abspath(self.KEYRING_FNAME)
        keyring_json_path = self.get_abspath(self.KEYRING_JSON_FNAME)

        keyring_tar_fname = '{}.tar.xz'.format(self.keyring_type)
        self.tar_path = self.get_abspath(keyring_tar_fname)

        logging.debug('Creating {}...'.format(self.tar_path))
        with tarfile.open(self.tar_path, 'w:xz') as fp:
            for abspath in [keyring_path, keyring_json_path]:
                logging.info('Adding {} to the tar file...'.format(abspath))
                fp.add(abspath, basename(abspath))

    def get_abspath(self, rel_path):
        """Get absolute path from a relative path.

        The path that is used as a base is the one in which the keyring files
        have been craeted.

        :param rel_path: Relative path to file/directory
        :type rel_path: str
        :returns: Absolute path to file/directory
        :rtype: str

        """
        return join(self.keyring_dir, rel_path)

    def sign(self, abspath):
        """Sign file.

        :param abspath: Absolute path to the file
        :type abspath: str
        :returns: Absolute path to the signature file (.asc)
        :rtype: str

        """
        logging.info('Signing {} with {} key...'
                     .format(abspath, self.keyring_type))
        with open(abspath, 'rb') as fp:
            signature = str(self.gpg.sign_file(fp, detach=True))

        signed_abspath = '{}.asc'.format(abspath)
        logging.debug('Writing signature to {}...'.format(signed_abspath))
        with open(signed_abspath, 'w') as fp:
            fp.write(signature)

        return signed_abspath

    def sign_keyring(self, other_keyring):
        """Sign keyring.

        :param other_keyring: Keyring object whose .tar file has to be signed
        :type other_keyrjing: upgrade.keyring.Keyring

        """
        logging.info('Signing {} keyring with {} key'
                     .format(other_keyring.keyring_type, self.keyring_type))
        other_keyring.asc_path = self.sign(other_keyring.tar_path)

    def export(self, target_dir):
        """Export keyring.

        Copy the .tar and .asc files to the target directory.

        :param target_dir: Directory where keyring files should be exported
        :type target_dir: str

        """
        logging.info('Exporting {} keyring...'.format(self.keyring_type))
        shutil.copy(self.tar_path, target_dir)
        shutil.copy(self.asc_path, target_dir)
