"""Easy generation of the files needed to run an upgrade test."""
import json
import hashlib
import logging
import os
import tarfile
import tempfile
import urllib.request

from os.path import dirname, join, relpath

from upgrade.environment import CLEANUP, JSON_OPTIONS


class UpgradeFiles(object):

    """Upgrade files generator object.

    :param device_obj: Device for which the upgrade files should be generated
    :type device_obj: upgrade.device.Device
    :param keyrings: Keyrings object that contains all the generated keyrings
    :type keyrings: upgrade.keyring.Keyrings

    """

    CHANNELS_FNAME = 'channels.json'
    INDEX_FNAME = 'index.json'

    CHANNELS = ['devel-proposed']
    DEVICES = ['mako']

    OFFICIAL_SERVER = 'http://system-image.ubuntu.com'

    def __init__(self, device_obj, keyrings):
        self.device_obj = device_obj
        self.keyrings = keyrings

        channel, device = device_obj.get_info()
        self.channel = channel
        self.device = device
        assert channel in self.CHANNELS
        assert device in self.DEVICES

        logging.info('Creating server files directory...')
        self.directory = tempfile.mkdtemp(prefix='upgrade-test-server-files')
        CLEANUP.add_directory(self.directory)

        logging.info('Creating gpg directory...')
        gpg_dir = join(self.directory, 'gpg')
        os.mkdir(gpg_dir)
        keyrings.export(gpg_dir)

        self._configure_files()

    def _configure_files(self):
        self._configure_channels_file()

        for channel in self.CHANNELS:
            for device in self.DEVICES:
                self._configure_index_file(
                    channel,
                    device,
                )

    def _configure_channels_file(self):
        logging.info('Creating {}...'.format(self.CHANNELS_FNAME))
        channels_data = {channel: {
            device: {'index': ('/{}/{}/{}'
                               .format(channel, device, self.INDEX_FNAME))}
            for device in self.DEVICES}
            for channel in self.CHANNELS
        }
        channels_path = join(self.directory, self.CHANNELS_FNAME)
        with open(channels_path, 'w') as fp:
            json.dump(channels_data, fp, **JSON_OPTIONS)

        self.keyrings['image-signing'].sign(channels_path)

    def _configure_index_file(self, channel, device):
        device_dir = join(self.directory, channel, device)
        os.makedirs(device_dir)

        # Get file from original server
        index_url = join(
            self.OFFICIAL_SERVER,
            channel,
            device,
            self.INDEX_FNAME)

        logging.info('Opening {}...'.format(index_url))
        response = urllib.request.urlopen(index_url)
        assert response.code == 200
        response_data = response.read().decode('utf-8')

        image_signing_keyring = self.keyrings['image-signing']
        index_data = json.loads(response_data)
        if channel == self.channel and device in self.device:
            self._configure_delta_file(
                index_data,
                device_dir,
                image_signing_keyring,
            )

        index_path = join(device_dir, self.INDEX_FNAME)
        with open(index_path, 'w') as fp:
            json.dump(index_data, fp, **JSON_OPTIONS)
        image_signing_keyring.sign(index_path)

    def _configure_delta_file(self, index_data, device_dir,
                              image_signing_keyring):
        delta_dir = tempfile.mkdtemp(prefix='upgrade-test-delta-')
        CLEANUP.add_directory(delta_dir)
        etc_dir = join(delta_dir, 'system', 'etc')
        os.makedirs(etc_dir)
        test_path = join(etc_dir, 'upgrade-test')
        with open(test_path, 'w') as fp:
            fp.write('Upgrade test')

        delta_fname = 'upgrade.tar.xz'
        delta_path = join(device_dir, delta_fname)
        with tarfile.open(delta_path, 'w:xz') as fp:
            fp.add(join(delta_dir, 'system'), 'system')

        with open(delta_path, 'rb') as fp:
            sha256 = hashlib.sha256()
            sha256.update(fp.read())
            checksum = sha256.hexdigest()

        delta_signature_path = \
            image_signing_keyring.sign(delta_path)

        images = index_data['images']
        last_full_image = next(
            image
            for image in reversed(images)
            if image['type'] == 'full'
        )
        last_version = last_full_image['version']

        index_data['images'].append({
            'description': 'UpgradeTest',
            'files': [{
                'checksum': checksum,
                'order': 0,
                'path': '/' + relpath(delta_path, self.directory),
                'signature': (
                    '/' + relpath(delta_signature_path, self.directory)),
                'size': os.path.getsize(delta_path),
            },
            ],
            'type': 'delta',
            'base': last_version,
            'version': last_version + 1,
        })

    def push(self):
        """Push upgrade test files to device."""
        base = os.path.join(dirname(__file__), 'files')
        target_base = '/tmp/upgrade-test'
        self.device_obj.push(base, target_base)
        self.device_obj.push(
            self.directory,
            join(target_base, 'server-files'))

        logging.debug(
            'Copying generated archive master key to default location...')
        self.device_obj.run(
            'cp '
            '/tmp/upgrade-test/server-files/gpg/archive-master.tar.* '
            '/etc/system-image')

        logging.debug('Patching ubuntu-download DBus service file')
        service_file = (
            '/usr/share/dbus-1/system-services/'
            'com.canonical.applications.Downloader.service')
        self.device_obj.run(
            r'sed -ri "s|^Exec=.*$|\0 -self-signed-certs '
            r'/tmp/upgrade-test/data/cert.pem|" {}'
            .format(service_file))
